class DisabledPipFlag(Exception):
    def __init__(self, message: str):
        self.message = message
        self.exit_code = 4001
        super().__init__(self.message)


class WrongPkgName(Exception):
    def __init__(self, message: str):
        self.message = message
        self.exit_code = 4002
        super().__init__(self.message)


class WrongSpecifierSet(Exception):
    def __init__(self, message: str):
        self.message = message
        self.exit_code = 4003
        super().__init__(self.message)


class EmptyDatabase(Exception):
    def __init__(self, message: str):
        self.message = message
        self.exit_code = 4004
        super().__init__(self.message)



import os
import sys
import responses
import pytest
from packaging.specifiers import Version
from pirg.exceptions import DisabledPipFlag, EmptyDatabase, WrongSpecifierSet, WrongPkgName
from pirg.utils import (
    PYPI_URL,
    check_for_pip_args,
    fuzzy_search,
    load_requirements_file,
    get_package,
    check_for_requirements_file,
    parse_package_name,
)


@pytest.fixture
def temporary_requirements_file(tmpdir):
    requirements_path = os.path.join(tmpdir, "requirements.txt")
    with open(requirements_path, "w") as req_file:
        req_file.write("package1==1.0.0\n")

    yield requirements_path

    os.remove(requirements_path)


def test_load_requirements_file(temporary_requirements_file):
    # test with file
    requirements = load_requirements_file(temporary_requirements_file)
    assert len(requirements) == 1
    assert "package1" in [pkg.name for pkg in requirements]

    # test empty file location
    requirements = load_requirements_file("")
    assert not requirements


def test_get_package():
    package_name = [
        "package1",
        "package2",
        "package3",
        "package4",
    ]

    specifier_set = ["", "", ">1.1", ">1.1"]
    mock_response_body = [
        {
            "releases": {
                "1.0.0": [{"requires_python": ">=3.8"}],
                "1.1.0": [{"requires_python": ">=3.8"}],
                "1.2.0": [{"requires_python": ">=3.8"}],
            }
        },
        {
            "releases": {
                "1.0.0": [{"requires_python": None}],
                "1.1.0": [{"requires_python": None}],
                "1.2.0": [{"requires_python": None}],
            }
        },
        {
            "releases": {
                "1.0.0": [{"requires_python": ">=3.8"}],
                "1.1.0": [{"requires_python": ">=3.8"}],
                "1.2.0": [{"requires_python": ">=3.8"}],
            }
        },
        {
            "releases": {
                "1.0.0": [{"requires_python": None}],
                "1.1.0": [{"requires_python": None}],
                "1.2.0": [{"requires_python": None}],
            }
        },
    ]

    for pkg, ss, mres in zip(package_name, specifier_set, mock_response_body):
        url = PYPI_URL(pkg)

        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, url, json=mres)
            result = get_package(pkg + ss)

            assert result.name == pkg
            assert Version("1.2.0") in result.specifier_set

    package_name = "package5"
    specifier_set = "==3.0"
    mock_response_body = {
        "releases": {
            "1.0.0": [{"requires_python": ">=3.8"}],
            "1.1.0": [{"requires_python": ">=3.8"}],
            "1.2.0": [{"requires_python": ">=3.8"}],
        }
    }
    url = PYPI_URL(package_name)

    with pytest.raises(WrongSpecifierSet):
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, url, json=mock_response_body)
            _ = get_package(package_name + specifier_set)


def test_check_for_requirements_file(tmpdir):
    root_dir = tmpdir.mkdir("project")
    sub_dir1 = root_dir.mkdir("subdirectory1")
    sub_dir2 = sub_dir1.mkdir("subdirectory2")

    sub_dir1.join("requirements.txt").write("Test requirements file content")

    # test if file is in sub_dir1
    os.chdir(str(sub_dir2))
    assert check_for_requirements_file() == sub_dir1.join("requirements.txt")

    # test if the current directory is default option if there is no file
    os.chdir(str(root_dir))
    assert check_for_requirements_file() == os.path.join(root_dir, "requirements.txt")


def test_check_for_pip_args(monkeypatch):
    test_argv = ["script_name", "arg1", "arg2", "--", "pip_arg1", "pip_arg2"]
    monkeypatch.setattr(sys, "argv", test_argv)
    result = check_for_pip_args()
    assert result == {"pip_arg1", "pip_arg2"}

    test_argv = ["script_name", "arg1", "arg2", "--"]
    monkeypatch.setattr(sys, "argv", test_argv)
    result = check_for_pip_args()
    assert not result

    test_argv = ["script_name", "arg1", "arg2"]
    monkeypatch.setattr(sys, "argv", test_argv)
    result = check_for_pip_args()
    assert not result

    test_argv = ["--"]
    monkeypatch.setattr(sys, "argv", test_argv)
    result = check_for_pip_args()
    assert not result

    test_argv = []
    monkeypatch.setattr(sys, "argv", test_argv)
    result = check_for_pip_args()
    assert not result

    test_argv = []
    monkeypatch.setattr(sys, "argv", test_argv)
    result = check_for_pip_args()
    assert not result

    test_argv = ["--", "-r", "requirements.txt"]
    monkeypatch.setattr(sys, "argv", test_argv)
    with pytest.raises(DisabledPipFlag):
        _ = check_for_pip_args()


def test_parse_package_name():
    package_name = "SomePackage[suffix]>=2.0.0,<=1.0.0"
    pkg_name, pkg_suffix, pkg_specifier_set = parse_package_name(package_name)
    assert pkg_name == "SomePackage"
    assert pkg_suffix == "suffix"
    assert pkg_specifier_set == ">=2.0.0,<=1.0.0"

    package_name = "[suffix]>=2.0.0,<=1.0.0"
    with pytest.raises(WrongPkgName):
        _, _, _ = parse_package_name(package_name)

    package_name = "SomePackage>=2.0.0"
    pkg_name, pkg_suffix, pkg_specifier_set = parse_package_name(package_name)
    assert pkg_name == "SomePackage"
    assert not pkg_suffix
    assert pkg_specifier_set == ">=2.0.0"

    package_name = "SomePackage"
    pkg_name, pkg_suffix, pkg_specifier_set = parse_package_name(package_name)
    assert pkg_name == "SomePackage"
    assert not pkg_suffix
    assert not pkg_specifier_set


def test_fuzzy_search():
    test_db = {
        "package1": "Package1",
        "package2": "Package2",
        "package3": "Package3",
    }

    search_term = "Package1"

    output = fuzzy_search(search_term, test_db)
    assert search_term in output

    search_term = ""
    output = fuzzy_search(search_term, test_db)
    assert not output

    search_term = "ASdwe"
    output = fuzzy_search(search_term, test_db)
    assert not output

    with pytest.raises(EmptyDatabase):
        _ = fuzzy_search(search_term, {})



from dataclasses import dataclass
from typing import Optional

from packaging.specifiers import SpecifierSet


@dataclass
class Package:
    name: str
    suffix: Optional[str] = None
    specifier_set: Optional[SpecifierSet] = None

    def __eq__(self, other) -> bool:
        return (
            self.name == other.name
            and self.specifier_set == other.specifier_set
            and self.suffix == other.suffix
        )

    def __hash__(self) -> int:
        return hash(self.name + str(self.suffix) + str(self.specifier_set))

    def __str__(self) -> str:
        string = f"{self.name}"
        string += f"[{self.suffix}]" if self.suffix else ""
        string += f"{str(self.specifier_set)}" if self.specifier_set else ""
        return string



import logging.config
import os
import subprocess
import sys
import tempfile
import traceback
from importlib import metadata
from typing import List

import typer
from requests.exceptions import HTTPError
from typing_extensions import Annotated

from pirg.config import log_config
from pirg.exceptions import (
    DisabledPipFlag,
    EmptyDatabase,
    WrongPkgName,
    WrongSpecifierSet,
)
from .models import Package
from .utils import (
    check_for_pip_args,
    check_for_requirements_file,
    check_if_pypi_simple_is_modified,
    create_db,
    create_requirements,
    fuzzy_search,
    get_package,
    get_pypi_simple_data,
    load_requirements_file,
    parse_package_name,
    run_subprocess,
)

TEMP_FILENAME = "pirg_pkg_db.txt"
__version__ = metadata.version("pirg")
logging.config.dictConfig(log_config)

main = typer.Typer()


def version_callback(value: bool):
    if value:
        typer.echo(f"{__version__}")
        raise typer.Exit()


@main.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(
        None,
        "--version",
        help="Current version",
        callback=version_callback,
    ),
):
    pass


@main.command()
def install(
    package_names: Annotated[List[str], typer.Argument(help="List of packages")] = None,
    requirements_path: Annotated[str, typer.Option()] = check_for_requirements_file(),
    update_all: Annotated[bool, typer.Option()] = False,
    log_level: Annotated[str, typer.Option(help="Set the log level")] = "INFO",
) -> None:
    """
    Installs [package_names] and puts them in the requirements file on [requirements_path] location

    You can pass additional `pip install` arguments after "--".

    Example:
        `pirg install torch -- --index-url https://download.pytorch.org/whl/cu118`

    """
    log_level = log_level.upper()
    log_level = getattr(logging, log_level)
    logging.getLogger().setLevel(log_level)
    logging.debug(f"argv: {sys.argv}")

    try:
        pip_args = check_for_pip_args()
        logging.debug(f"pip_args: {pip_args}")
        package_names = set(package_names) - pip_args

        current_pkgs = load_requirements_file(requirements_loc=requirements_path)
        new_pkgs = {get_package(package_name=pkg) for pkg in package_names}
        new_pkgs = new_pkgs - current_pkgs

        if update_all:
            # all pkgs update
            update_current_pkgs = {get_package(package_name=pkg.name) for pkg in current_pkgs}
            update_current_pkgs = new_pkgs.union(update_current_pkgs)
            new_pkgs = update_current_pkgs
        else:
            # one pkg update
            # check if it has same name but different version
            update_current_pkgs = {c for c in current_pkgs for n in new_pkgs if c.name != n.name}
            update_current_pkgs.update(new_pkgs)

        ins_pkgs = [f"{str(p)}" for p in new_pkgs]
        logging.debug(f"ins_pkgs: {ins_pkgs}")

        skip_pip_args = {"-h", "--help"}
        if not ins_pkgs and not update_all and not bool(skip_pip_args & pip_args):
            logging.info("Nothing to install")
            return

        run_subprocess(pkgs=ins_pkgs, pip_command="install", pip_args=list(pip_args))
        create_requirements(package_names=update_current_pkgs, requirements_loc=requirements_path)
    except FileNotFoundError as e:
        traceback.print_exc()
        sys.exit(e.errno)
    except HTTPError as e:
        pkg_name = e.args[0].split()[-1].split("/")[-2]
        logging.error(f"Failed to find the latest version of {pkg_name} on PyPI")
        traceback.print_exc()
        sys.exit(e.response.status_code)
    except subprocess.CalledProcessError as e:
        logging.error("Failed to install packages")
        traceback.print_exc()
        sys.exit(e.returncode)
    except (DisabledPipFlag, WrongPkgName, WrongSpecifierSet) as e:
        logging.error(str(e))
        sys.exit(e.exit_code)


@main.command()
def uninstall(
    package_names: Annotated[List[str], typer.Argument()] = None,
    requirements_path: Annotated[str, typer.Option()] = check_for_requirements_file(),
    delete_all: Annotated[bool, typer.Option()] = False,
    log_level: Annotated[str, typer.Option(help="Set the log level")] = "INFO",
) -> None:
    """
    Uninstalls [package_names] and removes them from the requirements file on [requirements_path] location

    You can pass additional `pip uninstall` arguments after "--".

    Example:
        `pirg uninstall torch -- --yes`

    """
    log_level = log_level.upper()
    log_level = getattr(logging, log_level)
    logging.getLogger().setLevel(log_level)
    logging.debug(f"argv: {sys.argv}")

    try:
        pip_args = check_for_pip_args()
        logging.debug(f"pip_args: {pip_args}")
        package_names = set(package_names) - pip_args

        package_names = [parse_package_name(val) for val in package_names]
        new_pkgs = set([Package(n, sf, ss) for n, sf, ss in package_names])
        current_pkgs = load_requirements_file(requirements_loc=requirements_path)

        if delete_all:
            new_pkgs.update(current_pkgs)

        # repopulate version from requirements.txt
        new_pkgs = {c for c in current_pkgs for n in new_pkgs if c.name == n.name}
        current_pkgs = current_pkgs - new_pkgs

        rm_pkgs = [p.name for p in new_pkgs]

        skip_pip_args = {"-h", "--help"}
        if not rm_pkgs and not delete_all and not bool(skip_pip_args & pip_args):
            logging.info("Nothing to remove")
            return

        run_subprocess(pkgs=rm_pkgs, pip_command="uninstall", pip_args=list(pip_args))
        create_requirements(package_names=current_pkgs, requirements_loc=requirements_path)
    except FileNotFoundError as e:
        traceback.print_exc()
        sys.exit(e.errno)
    except HTTPError as e:
        pkg_name = e.args[0].split()[-1].split("/")[-2]
        logging.error(f"Failed to find the latest version of {pkg_name} on PyPI")
        traceback.print_exc()
        sys.exit(e.response.status_code)
    except subprocess.CalledProcessError as e:
        logging.error("Failed to remove packages")
        traceback.print_exc()
        sys.exit(e.returncode)
    except (DisabledPipFlag, WrongPkgName, WrongSpecifierSet) as e:
        logging.error(str(e))
        sys.exit(e.exit_code)


@main.command()
def search(
    user_input: Annotated[str, typer.Argument()] = None,
    log_level: Annotated[str, typer.Option(help="Set the log level")] = "INFO",
) -> None:
    """
    Search for python package on PYPI

    Example:
        `pirg search sqlalchemy` -> Search result: ['SQLAlchemy', 'sqlalchemyp',...]
    """
    log_level = log_level.upper()
    log_level = getattr(logging, log_level)
    logging.getLogger().setLevel(log_level)
    logging.debug(f"argv: {sys.argv}")

    try:
        temp_dir = tempfile.gettempdir()
        filename = os.path.join(temp_dir, TEMP_FILENAME)

        if not os.path.exists(filename):
            raise FileNotFoundError("Package names file doesn't exist. Please run `initdb` first.")

        if check_if_pypi_simple_is_modified():
            # fmt: off
            logging.info("Current list of package names is out of date. Please update with `initdb --update`")
            # fmt: on

        with open(filename, "r") as file:
            package_names = [line.strip() for line in file]

        indexed_package_names = {name.lower(): name for name in package_names}
        search_output = fuzzy_search(user_input, indexed_package_names)
        logging.info(f"Search result: {search_output}")
    except EmptyDatabase as e:
        logging.error(str(e))
        sys.exit(e.exit_code)
    except FileNotFoundError as e:
        traceback.print_exc()
        sys.exit(e.errno)
    except HTTPError as e:
        logging.error(e)
        traceback.print_exc()
        sys.exit(e.response.status_code)


@main.command()
def initdb(
    update: Annotated[bool, typer.Option()] = False,
    log_level: Annotated[str, typer.Option(help="Set the log level")] = "INFO",
) -> None:
    """
    Initialize or update current package names list

    Example:
        `pirg initdb`
    """
    log_level = log_level.upper()
    log_level = getattr(logging, log_level)
    logging.getLogger().setLevel(log_level)
    logging.debug(f"argv: {sys.argv}")

    try:
        temp_dir = tempfile.gettempdir()
        filename = os.path.join(temp_dir, TEMP_FILENAME)
        logging.debug(f"Database location: {filename}")

        if not update and os.path.exists(filename):
            logging.info("Database already initialized")
            return

        new_version = check_if_pypi_simple_is_modified()
        if update and not new_version:
            logging.info("Database is up-to-date")
            return

        logging.info("Downloading data")
        data = get_pypi_simple_data()

        create_db(filename, data)
        logging.info("Database initialized")
    except FileNotFoundError as e:
        traceback.print_exc()
        sys.exit(e.errno)
    except HTTPError as e:
        logging.error(e)
        traceback.print_exc()
        sys.exit(e.response.status_code)


if __name__ == "__main__":
    main()



import logging
import os
import re
import subprocess
import sys
from datetime import datetime, timedelta
from difflib import get_close_matches
from typing import Dict, List, Optional, Set, Tuple

import requests
from bs4 import BeautifulSoup
from fuzzywuzzy import fuzz, process
from packaging.specifiers import Specifier, SpecifierSet
from packaging.version import Version

from .exceptions import DisabledPipFlag, WrongPkgName, WrongSpecifierSet, EmptyDatabase
from .models import Package

PYPI_URL = lambda pkg_name: f"https://pypi.org/pypi/{pkg_name}/json"
PYPI_SIMPLE_URL = "https://pypi.org/simple/"
PY_VERSION = Version(sys.version.split()[0])
PARSE_PATTERN = r"^(?P<name>[a-zA-Z0-9_-]+)(\[(?P<suffix>[a-zA-Z0-9_-]+)\])?(?P<specifier_set>.*)"
REQUIREMENTS = "requirements.txt"


def parse_package_name(pkg: str) -> Tuple[str, Optional[str], Optional[str]]:
    pattern = PARSE_PATTERN

    match = re.match(pattern, pkg)
    if not match:
        raise WrongPkgName(f"Package {pkg} does not match pattern")

    package_data = match.groupdict()

    name = package_data["name"]
    suffix = package_data["suffix"]
    specifier_set = package_data["specifier_set"]

    return name, suffix, specifier_set


def create_requirements(
    package_names: Set[Package],
    requirements_loc: str,
    flag: str = "w",
) -> None:
    with open(requirements_loc, flag) as req_file:
        for pkg in package_names:
            req_file.write(f"{str(pkg)}\n")


def load_requirements_file(requirements_loc: str) -> Set[Package]:
    requirements = set()

    if not os.path.exists(requirements_loc):
        return requirements

    with open(requirements_loc, "r") as req_file:
        for line in req_file:
            name, suffix, specifier_set = parse_package_name(line)
            pkg = Package(
                name=name,
                suffix=suffix,
                specifier_set=specifier_set,
            )
            requirements.add(pkg)

    return requirements


def get_package(package_name: str) -> Package:
    pkg_name, pkg_suffix, pkg_specifier_set = parse_package_name(package_name)

    response = requests.get(PYPI_URL(pkg_name=pkg_name))
    response.raise_for_status()
    package_data = response.json()

    valid_versions: Set[Version] = {
        Version(rel)
        for rel in package_data["releases"]
        for elem in package_data["releases"][rel]
        if elem["requires_python"] is not None
        and PY_VERSION in SpecifierSet(elem["requires_python"])
    }

    logging.debug(f"valid_versions: {valid_versions}")
    if not valid_versions:
        # when `requires_python = None` for all pkgs
        valid_versions = {Version(rel) for rel in package_data["releases"]}

    pkg_specifier_set = SpecifierSet(pkg_specifier_set) if pkg_specifier_set else None
    logging.debug(f"pkg_specifier_set: {pkg_specifier_set}")

    if pkg_specifier_set:
        specifier_set = {pkg_specifier_set for vv in valid_versions if vv in pkg_specifier_set}

        if not specifier_set:
            # if the specifier is wrong we get from packaging lib Invalid specifier
            # this means, here can only be empty specifier set
            # which can happen if valid version is not in provided specifier set
            raise WrongSpecifierSet(f"Not valid specifier set: {pkg_specifier_set}")

        specifier_set = specifier_set.pop()
    else:
        # do not allow pre, post or dev releases
        valid_versions = {
            v
            for v in valid_versions
            if not (v.is_prerelease or v.is_postrelease or v.is_devrelease)
        }
        max_version = max(valid_versions)
        specifier_set = SpecifierSet(f"=={max_version}")

    logging.debug(f"specifier_set: {specifier_set}")
    return Package(name=pkg_name, suffix=pkg_suffix, specifier_set=specifier_set)


def check_for_pip_args() -> Set[str]:
    try:
        dash_idx = sys.argv.index("--") + 1
        pip_args = set(sys.argv[dash_idx:])
    except ValueError:
        pip_args = set()

    disable = {"-r", "--requirements", "-U", "--upgrade"}
    disable = disable & pip_args
    if disable:
        raise DisabledPipFlag(f"{disable} is disabled")

    return pip_args


def check_for_requirements_file() -> str:
    current_dir = os.getcwd()
    while True:
        if REQUIREMENTS in os.listdir(current_dir):
            return os.path.join(current_dir, REQUIREMENTS)

        parent_dir = os.path.dirname(current_dir)
        if parent_dir == current_dir:
            break

        current_dir = parent_dir

    return os.path.join(os.getcwd(), REQUIREMENTS)


def get_pypi_simple_data(url: str = PYPI_SIMPLE_URL) -> str:
    response = requests.get(url)
    response.raise_for_status()
    return response.text


def check_if_pypi_simple_is_modified(days: int = 3, url: str = PYPI_SIMPLE_URL) -> bool:
    current_date = datetime.now()
    last_modified_date = current_date - timedelta(days=days)
    if_modified_since = last_modified_date.strftime("%a, %d %b %Y %H:%M:%S GMT")

    headers = {"If-Modified-Since": if_modified_since}

    response = requests.head(url, headers=headers)
    response.raise_for_status()

    if response.status_code == 200:
        return True
    elif response.status_code == 304:
        return False


def create_db(filename: str, data: str) -> None:
    soup = BeautifulSoup(data, "html.parser")
    links = soup.find_all("a")
    package_names = [link.text.strip() for link in links if link.text.strip()]

    with open(filename, "w") as file:
        for package in package_names:
            file.write(package + "\n")


def fuzzy_search(search_input: str, indexed_pkg_names: Dict[str, str]) -> List[str]:
    if not indexed_pkg_names:
        raise EmptyDatabase("Empty DB")

    matches = get_close_matches(search_input, indexed_pkg_names, n=7, cutoff=0.6)
    search_results = process.extract(search_input, matches, scorer=fuzz.ratio)

    org_names = [indexed_pkg_names[result] for result, _ in search_results]
    return org_names


def run_subprocess(pkgs: List[str], pip_command: str, pip_args: List[str]):
    subprocess.run(["pip", pip_command] + pkgs + pip_args, check=True)
    logging.info(f"{pip_command.capitalize()}ed packages: {pkgs}")



from rich.logging import RichHandler

log_config = {
    "version": 1,
    "formatters": {
        "verbose": {
            "format": "%(message)s",
            "datefmt": "[pirg]",
        },
    },
    "handlers": {
        "console": {
            "class": "rich.logging.RichHandler",
            "formatter": "verbose",
        },
    },
    "root": {
        "handlers": ["console"],
    },
}



import logging
import os.path
import sys
import requests
import pytest
from requests import HTTPError
from pirg.pirg import initdb, install, uninstall, search
from pirg.pirg import TEMP_FILENAME

# TODO: test update all
# FIXME: try to mock packages
# FIXME: mock run_subprocess


def mock_get_package(package_name):
    status_code = 404
    response = requests.Response()
    response.status_code = status_code

    raise HTTPError(response=response)


def test_install(tmpdir, monkeypatch, caplog):
    test_dir = tmpdir.mkdir("test_dir")
    package_names = ["python-dotenv", "numpy"]
    requirements_file = test_dir.join("requirements.txt")
    caplog.set_level(logging.INFO)

    monkeypatch.setattr("pirg.utils.load_requirements_file", lambda x: set())
    monkeypatch.setattr("pirg.utils.run_subprocess", lambda pkgs, cmd, args: None)

    # check if requirements exist
    install(package_names=package_names, requirements_path=requirements_file.strpath)
    assert requirements_file.exists()

    # test installing already installed packages
    install(package_names=package_names, requirements_path=requirements_file.strpath)
    assert "Nothing to install" in [rec.message for rec in caplog.records]

    # test installing with no packages specified (nothing to install)
    install(package_names=[], requirements_path=requirements_file.strpath)
    assert "Nothing to install" in [rec.message for rec in caplog.records]

    # test installing with disabled pip arguments but no packages specified
    monkeypatch.setattr(sys, "argv", ["--", "-U"])
    with pytest.raises(SystemExit) as excinfo:
        install(package_names=[], requirements_path=requirements_file.strpath)
    assert excinfo.value.code == 4001
    monkeypatch.setattr(sys, "argv", [])

    # test wrong package name
    package_names = ["package1"]
    monkeypatch.setattr("pirg.utils.get_package", mock_get_package)
    with pytest.raises(SystemExit) as excinfo:
        install(package_names=package_names, requirements_path=requirements_file.strpath)
    assert excinfo.value.code == 404

    # test installing with no args
    install(package_names=[])
    assert "Nothing to install" in [rec.message for rec in caplog.records]


def test_uninstall(tmpdir, monkeypatch, caplog):
    test_dir = tmpdir.mkdir("test_dir")
    package_names = ["python-dotenv", "numpy"]
    requirements_file = test_dir.join("requirements.txt")
    caplog.set_level(logging.INFO)

    monkeypatch.setattr("pirg.utils.load_requirements_file", lambda x: set())
    monkeypatch.setattr("pirg.utils.run_subprocess", lambda pkgs, cmd, args: None)

    # create requirements file
    install(package_names=package_names, requirements_path=requirements_file.strpath)
    monkeypatch.setattr(sys, "argv", ["--", "-y"])

    # uninstalling already installed packages
    uninstall(package_names=package_names, requirements_path=requirements_file.strpath)
    assert os.path.getsize(requirements_file) == 0

    # uninstalling already uninstalled packages
    uninstall(package_names=package_names, requirements_path=requirements_file.strpath)
    assert "Nothing to remove" in [rec.message for rec in caplog.records]

    # test uninstalling all from requirements
    monkeypatch.setattr(sys, "argv", [])
    install(package_names=package_names, requirements_path=requirements_file.strpath)
    monkeypatch.setattr(sys, "argv", ["--", "-y"])
    uninstall(package_names=[], requirements_path=requirements_file.strpath, delete_all=True)
    assert os.path.getsize(requirements_file) == 0

    # test uninstalling with no packages specified (nothing to uninstall)
    uninstall(package_names=[], requirements_path=requirements_file.strpath)
    assert "Nothing to remove" in [rec.message for rec in caplog.records]

    # test uninstalling with pip disabled arguments but no packages specified
    monkeypatch.setattr(sys, "argv", ["--", "-r", f"{requirements_file.strpath}"])
    with pytest.raises(SystemExit) as excinfo:
        uninstall(package_names=[], requirements_path=requirements_file.strpath)
    assert excinfo.value.code == 4001
    monkeypatch.setattr(sys, "argv", [])

    # test wrong package name
    package_names = ["package1"]
    uninstall(package_names=package_names, requirements_path=requirements_file.strpath)
    assert "Nothing to remove" in [rec.message for rec in caplog.records]


def test_initdb(monkeypatch, tmpdir, caplog):
    def mock_data():
        return "package1\npackage2\npackage3"

    caplog.set_level(logging.INFO)

    monkeypatch.setattr("tempfile.gettempdir", lambda: tmpdir.strpath)
    monkeypatch.setattr(
        "pirg.utils.check_if_pypi_simple_is_modified.__code__",
        (lambda: False).__code__,
    )
    monkeypatch.setattr(
        "pirg.utils.get_pypi_simple_data.__code__",
        mock_data.__code__,
    )

    # default
    initdb()
    assert "Database initialized" in [rec.message for rec in caplog.records]

    # already initialized
    initdb()
    assert "Database already initialized" in [rec.message for rec in caplog.records]

    # update
    initdb(update=True)
    assert "Database is up-to-date" in [rec.message for rec in caplog.records]


def test_search(tmpdir, monkeypatch, caplog):
    package_names = ["package1", "paCKage2", "Package3"]
    filename = os.path.join(tmpdir.strpath, TEMP_FILENAME)
    with open(filename, "w") as file:
        for package in package_names:
            file.write(package + "\n")

    caplog.set_level(logging.INFO)

    monkeypatch.setattr("tempfile.gettempdir", lambda: tmpdir.strpath)
    monkeypatch.setattr(
        "pirg.utils.check_if_pypi_simple_is_modified.__code__",
        (lambda: False).__code__,
    )

    user_input = "package3"
    search(user_input)
    assert "Search result: ['Package3', 'paCKage2', 'package1']" in [
        rec.message for rec in caplog.records
    ]

    user_input = ""
    search(user_input)
    assert "Search result: []" in [rec.message for rec in caplog.records]

    with pytest.raises(TypeError):
        search()

    monkeypatch.setattr(
        "pirg.utils.check_if_pypi_simple_is_modified.__code__",
        (lambda: True).__code__,
    )
    user_input = ""
    search(user_input)
    assert (
        "Current list of package names is out of date. Please update with `initdb --update`"
        in [rec.message for rec in caplog.records]
    )

    filename = os.path.join(tmpdir.strpath, TEMP_FILENAME)
    with open(filename, "w") as file:
        file.write("")

    user_input = ""
    with pytest.raises(SystemExit) as excinfo:
        search(user_input)
    assert excinfo.value.code == 4004
